/*
 *  textProcessor.cpp
 *  lexical
 *
 *  Created by Chris Harrison on 10/29/05.
 *  Copyright 2005 __MyCompanyName__. All rights reserved.
 *
 */

#include "textProcessor.h"
#include "parseUtilities.h"
#include <string.h>
#include <iostream>

char * text;
bool previousTokenWasInteger;

void process(char * textPtr)
{
	text = textPtr;
	
	char * token;
	
	while (isEOF(text[0])==false)
	{	
		
		//printf("Next char %c. ", text[0]); //show next character to be checked
		//printf("\n"); //seperate tokens in output
		
		
		//non-tokenizable lexical classes
		//-------------------------------
		if (text[0]=='"') //found quote - string - special case
		{
			if (checkAndGetString())
			{
				printf(" <STRING> ");
				continue;
			}
			else
			{
				printf(" error on string ");
				exit (-1);
			}
		}
		
		
		if (text[0]=='/' && text[1]=='/') //found double slash - comment - special case
		{
			if (checkAndGetComment())
			{
				//printf(" <COMMENT> ");
				continue;
			}
			else
			{
				printf(" error on comment ");
				exit (-1);
			}
		}
		
		
		if (text[0]=='\'')  //found single quote - char - special case
		{
			if (getChar())
			{
				printf(" <CHAR> ");
				continue;
			}
			else
			{
				printf(" error on char ");
				exit (-1);
			}
		}
		
		if (text[0]==':' && text[1]=='=')
		{
			printf(":=");
			printf(" <OPERATION> ");
			text+=2;		
			continue;
		}
		
		if (text[0]=='.' && text[1]=='.')
		{
			printf("..");
			printf(" <KEYWORD> ");
			text+=2;		
			continue;
		}
		
		if (isSymbol(text[0])) //found symbol - symbol - special case
		{
			printf("%c",text[0]);
			printf(" <SYMBOL> ");
			text++;		
			continue;
		}
		
		
		//tokenizable lexical classes
		//---------------------------
		
		token = getNextToken(); //get the next token
		printf("%s", token); //output token
		
		if (isKeyword(token)) //is token a keyword
		{
			printf(" <KEYWORD> ");
			continue;
		}
		
		
		if (isOperation(token)) //is token an operation
		{
			printf(" <OPERATION> ");
			continue;
		}
		
		if (isLetter(token[0])) //is token a letter (possibily the start of an ID)
		{
			//printf("found letter \n");
			if (isID(token)) //is token an ID
			{
				printf(" <ID> ");
				continue;
			}
			else
			{
				printf(" error on ID ");
				exit (-1);
			}
		}
		
		if (isDigit(token[0]) || (token[0]=='-' && isDigit(token[1])))  //is token a numerical digit (possibily the start of a longer number)
		{
			//printf("found digit \n");
			if (isInt(token)) //is token an integer
			{
				printf(" <INTEGER> ");	
				continue;
			}
			else
			{
				printf(" error on integer");
				exit (-1);
			}
		}

			
		
		
	}	//end loop
	
} //end process function



//gets the next token from the global text pointer
//***manipulates text pointer***
char* getNextToken()
{
	char * tempPtr = text; //save pointer to text
	char * newToken;
	int count;
	
	bool skippedWhitespace = false; 
	bool skippedText = false;
	
	count=0;
	while (isWhitespace(text[0])==false && isEOF(text[0])==false && isSymbol(text[0])==false) //keep reading text until whitespace or symbol is reached
	{
		//printf("found %c \n", text[0]);
		skippedText=true;
		text+=1; //advance to next character
		count++;
	}
	
	if (skippedText==true) 
		//printf("Skipped %i characters of text \n", count);
	
	if (isSymbol(text[0])==true)
	{
		newToken = (char*)malloc(count+1);
		memcpy(newToken,tempPtr,count);
		newToken[count+1]=0;
		
		return newToken;
	}
		
	
		
		
	count=0;
	while (isWhitespace(text[0])==true) //keep reading whitespace until text is reached
	{
		//printf("found %c \n", text[0]);
		skippedWhitespace=true;
		text[0]=0; //zero out whitespace to delimit tokens
		text+=1; //advance to next character
		count++;
	}
	
	if (skippedWhitespace==true)
	{
		//printf(" <WHITESPACE> ");  //output identification of whitespace
		//printf("Skipped %i characters of whitespace \n", count);
	}
	return tempPtr;
}


//check to see if next token is a valid character
//***manipulates text pointer***
bool getChar()
{
	char * start = text; //save pointer front of string
	
	text++; //skip past opening single quote
		
	if (text[0]=='\'' && text[0]!='\'') //empty char constant, error
		return false;
	
	
	//check if current caracter not escape, and next character is single quote. 
	//also handles ''', '"', and general cases ('x')
	if (text[1]=='\'' && text[0]!='\\') //check cases without escape character
	{
		text+=2; //skip over character and ending quote
		text[0]=0; //zero out string to terminate token
		text++; //move to begining of next string
		printf("%s", start); //print string
		return true;
	}
	
	
	if (text[0]=='\\')
	{
		text++;
		
		if (isLetter(text[0]) && text[1]=='\'') // '\x' case (x any letter)
		{
			//printf("is letter? \n\n");
			text+=2; //skip over character and ending quote
			text[0]=0; //zero out string to terminate token
			text++; //move to begining of next string
			printf("%s", start); //print string
			return true;
		}
		
		else if (text[0]=='\"' && text[1]=='\'') // '\"' special case
		{
			text+=2; //skip over character and ending quote
			text[0]=0; //zero out string to terminate token
			text++; //move to begining of next string
			printf("%s", start); //print string
			return true;
		}
		
		else if (text[0]=='\'' && text[1]=='\'') // '\'' special case
		{
			text+=2; //skip over character and ending quote
			text[0]=0; //zero out string to terminate token
			text++; //move to begining of next string
			printf("%s", start); //print string
			return true;
		}
		
		else if (text[0]=='\\' && text[1]=='\'') // '\\' special case
		{
			text+=2; //skip over character and ending quote
			text[0]=0; //zero out string to terminate token
			text++; //move to begining of next string
			printf("%s", start); //print string
			return true;
		}
		
		else if (isDigit(text[0])==true) //octal digit case
		{
			int i = 0;
			
			while (isDigit(text[0+i]) && i<4) //check how many digits are escaped
			{
				i++;
			}		
			
			if (i>=4) //more than three digits, error
				return false;
			
			else //3 or less digits
			{
				text+=i; //advance text pointer over digits
				
				if (text[0]=='\'') //make sure closing single quote is in place
				{
					text++; //advance over closing single quote
					text[0]=0; //zero out string to terminate token
					text++; //move to begining of next string
					printf("%s", start); //print string
					return true;
				}
				else
					return false; //no closing single quote or more trailing characters! Error
				
			}
		}
	}
	
	//Ffall through case.  Shouldn't get here.  Error
	return false;	
	
}

//Check to see if the next token is a string
//***manipulates text pointer***
bool checkAndGetString()
{
	char * start = text; //save pointer front of string
	
	text++; //skip past opening quote
	
	if (text[0]=='"') //empty string case
		return true;	
	
	
	while (1 && isEOF(text[0])==false) //continue unless we reach end of file
	{
		if (text[0]=='"' && text[-1]!='\\') //if we find quote that is not escaped, found end of string!  
			break; //break out of loop
		
		text++; //otherwise, advance text pointer to next character
	}
	
	if (isEOF(text[0])==true) //reached end of file.  Probably no closing quote!  Error
		return false;
	
	text++; //advance text character past closing quote
	text[0]=0; //zero out string to terminate token
	text++; //move to begining of next string
	printf("%s", start); //print quote
	
	return true;
}


//check to see if next token is a comment
//***manipulates text pointer***
bool checkAndGetComment()
{
	char * start = text; //save pointer front of string
	
	text+=2; //skip past opening double slashes (//)
	
	
	if (isEndOfLine(text[0])) //check to see if end of line is directly after double slash
	{
		//empty comment case
		return true;
	}
	
	
	while (1 && isEOF(text[0])==false)  //continue unless we reach end of file
	{
		//printf("found %c \n", text[0]);
		
		if (isEndOfLine(text[0]))
			break;
		
		text++;
	}
	
	
	text[0]=0; //zero out end of line character
	text++; //move to begining of next string
	//printf("%s", start); //print comment
	
	return true;
}
